package interpreteur.parser.grammatical;

import interpreteur.parser.SyntaxException;
import interpreteur.parser.lexical.MotLexical;
import interpreteur.logique.ExpressionAbstraite;
import interpreteur.logique.Reference;
import interpreteur.logique.expressions.nonterminales.PrincipaleExpression;
import interpreteur.logique.expressions.terminales.Addition;
import interpreteur.logique.expressions.terminales.Affectation;
import interpreteur.logique.expressions.terminales.Different;
import interpreteur.logique.expressions.terminales.Division;
import interpreteur.logique.expressions.terminales.Egal;
import interpreteur.logique.expressions.terminales.Et;
import interpreteur.logique.expressions.terminales.Inferieur;
import interpreteur.logique.expressions.terminales.InferieurStrict;
import interpreteur.logique.expressions.terminales.Multiplication;
import interpreteur.logique.expressions.terminales.Nombre;
import interpreteur.logique.expressions.terminales.Non;
import interpreteur.logique.expressions.terminales.Ou;
import interpreteur.logique.expressions.terminales.Soustraction;
import interpreteur.logique.expressions.terminales.Superieur;
import interpreteur.logique.expressions.terminales.SuperieurStrict;
import interpreteur.logique.resultats.InvalidInstruction;
import interpreteur.logique.resultats.ResultatBooleen;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * Le parser grammatical valide la grammaire d'une séquence de mots
 * lexicaux et transorme cette séquence en une expression interprétable
 * 
 * @author François Biollet, Guillaume Delhumeau & Loïc Geoffroy
 */
public class ParserGrammatical
{

	private Map<Integer, Etat> correspondanceTypeLexicalEtat;
	private Etat etatDebut;
	private Etat etatNombre;
	private Etat etatVariable;
	private Etat etatOperation;
	private Etat etatNon;
	private Etat etatAffectation;
	private Etat etatParentheseOuvrante;
	private Etat etatParentheseFermante;
	private Etat etatFin;
	private boolean debug;

	public ParserGrammatical()
	{
		/*
		 * Dans un premier temps, on créé l'automate de validation
		 */
		correspondanceTypeLexicalEtat = new HashMap<Integer, Etat>();

		etatDebut = new Etat();
		etatNombre = new Etat();
		etatVariable = new Etat();
		etatOperation = new Etat();
		etatNon = new Etat();
		etatAffectation = new Etat();
		etatFin = new Etat();
		etatParentheseOuvrante = new Etat();
		etatParentheseFermante = new Etat();

		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_NOMBRE, etatNombre);
		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_BOOLEEN, etatNombre);
		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_OPERATEUR, etatOperation);
		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_NON, etatNon);
		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_FIN_INSTRUCTION, etatFin);
		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_VARIABLE, etatVariable);
		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_AFFECTATION, etatAffectation);
		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_PARENTHESE_OUVRANTE, etatParentheseOuvrante);
		correspondanceTypeLexicalEtat.put(MotLexical.TYPE_PARENTHESE_FERMANTE, etatParentheseFermante);

		etatDebut.addEtatAccessible(etatNombre);
		etatDebut.addEtatAccessible(etatVariable);
		etatDebut.addEtatAccessible(etatNon);
		etatDebut.addEtatAccessible(etatFin);
		etatDebut.addEtatAccessible(etatParentheseOuvrante);

		etatNombre.addEtatAccessible(etatOperation);
		etatNombre.addEtatAccessible(etatFin);
		etatNombre.addEtatAccessible(etatParentheseFermante);

		etatVariable.addEtatAccessible(etatOperation);
		etatVariable.addEtatAccessible(etatFin);
		etatVariable.addEtatAccessible(etatAffectation);
		etatVariable.addEtatAccessible(etatParentheseFermante);

		etatNon.addEtatAccessible(etatVariable);
		etatNon.addEtatAccessible(etatNombre);
		etatNon.addEtatAccessible(etatParentheseOuvrante);

		etatOperation.addEtatAccessible(etatVariable);
		etatOperation.addEtatAccessible(etatNombre);
		etatOperation.addEtatAccessible(etatNon);
		etatOperation.addEtatAccessible(etatParentheseOuvrante);

		etatAffectation.addEtatAccessible(etatNon);
		etatAffectation.addEtatAccessible(etatVariable);
		etatAffectation.addEtatAccessible(etatNombre);
		etatAffectation.addEtatAccessible(etatParentheseOuvrante);

		etatParentheseOuvrante.addEtatAccessible(etatNon);
		etatParentheseOuvrante.addEtatAccessible(etatVariable);
		etatParentheseOuvrante.addEtatAccessible(etatNombre);
		etatParentheseOuvrante.addEtatAccessible(etatParentheseFermante);
		etatParentheseOuvrante.addEtatAccessible(etatParentheseOuvrante);

		etatParentheseFermante.addEtatAccessible(etatOperation);
		etatParentheseFermante.addEtatAccessible(etatFin);
		etatParentheseFermante.addEtatAccessible(etatParentheseFermante);

		debug = false;
	}

	/**
	 * Analyse une séquence de mots lexicaux et retourne cette séquence
	 * sous le forme d'une Expression à interpréter
	 * @param seq
	 * @return
	 * @throws interpreteur.interfaces.parser.SyntaxException
	 */
	public ExpressionAbstraite analyse(ArrayList<MotLexical> seq) throws SyntaxException, InvalidInstruction
	{
		if (debug)
		{
			System.out.println("== Analyse grammaticale ==");
		}

		validerGrammaire(seq);
		ExpressionAbstraite exp = transformerEnExpressions(seq);
		return exp;
	}

	/**
	 * Validation de la grammaire avec un automate
	 * @param seq Liste des séquences à valider
	 * @throws interpreteur.interfaces.parser.SyntaxException Erreur de syntaxe
	 */
	private void validerGrammaire(ArrayList<MotLexical> seq) throws SyntaxException
	{
		/*
		 * On utilise l'automate fini créé dans le constructeur
		 */
		Etat etatCourant = etatDebut;
		for (int i = 0; i < seq.size(); ++i)
		{
			MotLexical motCourant = seq.get(i);
			etatCourant = etatCourant.changerEtat(motCourant.getValeur(),
							correspondanceTypeLexicalEtat.get(motCourant.getType()));

		}
	}

	/**
	 * Transforme une séquences de mots lexicaux en expressions à interpréter
	 * @param seq
	 * @return
	 * @throws interpreteur.interfaces.parser.SyntaxException
	 */
	private ExpressionAbstraite transformerEnExpressions(ArrayList<MotLexical> seq)
					throws SyntaxException, InvalidInstruction
	{

		ArrayList<ExpressionAbstraite> listeExpressions =
						new ArrayList<ExpressionAbstraite>();

		int compteurParenthese = 0;
		// Boucle de transformation
		for (int i = 0; i < seq.size(); i++)
		{
			MotLexical motCourant = seq.get(i);
			if (motCourant.getType() == MotLexical.TYPE_PARENTHESE_OUVRANTE)
			{
				compteurParenthese++;
			}
			else if (motCourant.getType() == MotLexical.TYPE_PARENTHESE_FERMANTE)
			{
				compteurParenthese--;
				if (compteurParenthese < 0)
				{
					throw new InvalidInstruction("Il y a une parenthèse fermante de trop.");
				}

			}
			else
			{
				if (i == seq.size() - 1 && compteurParenthese > 0)
				{
					throw new InvalidInstruction("Il manque une parenthèse fermante.");
				}
				ExpressionAbstraite exp = getExpression(motCourant);
				exp.setPriorite(exp.getPriorite() + compteurParenthese * 100);
				listeExpressions.add(exp);
			}
		}

		completerLesExpressions(listeExpressions);
		ExpressionAbstraite expFinale = listeExpressions.get(0);

		if (debug)
		{
			System.out.println("Expression finale : " + expFinale);
		}
		return expFinale;
	}

	/**
	 * Générer une Expression à partir d'un mot lexcial
	 * @param mot Le MotLexical à transformer
	 * @return l'expression correspondante
	 */
	private ExpressionAbstraite getExpression(MotLexical mot) throws SyntaxException
	{

		/*
		 * On teste chaque mot existant, tout bêtement
		 */
		if (mot.getType() == MotLexical.TYPE_NOMBRE)
		{
			return new Nombre(Double.parseDouble(mot.getValeur()));
		}
		if (mot.getType() == MotLexical.TYPE_BOOLEEN)
		{
			if (mot.getValeur().equals("vrai"))
			{
				return new ResultatBooleen(true);
			}
			else
			{
				return new ResultatBooleen(false);
			}
		}
		else if (mot.getType() == MotLexical.TYPE_OPERATEUR)
		{
			if (mot.getValeur().equals("+"))
			{
				return new Addition(null, null);
			}
			else if (mot.getValeur().equals("-"))
			{
				return new Soustraction(null, null);
			}
			else if (mot.getValeur().equals("*"))
			{
				return new Multiplication(null, null);
			}
			else if (mot.getValeur().equals("/"))
			{
				return new Division(null, null);
			}
			else if (mot.getValeur().equals(">="))
			{
				return new Superieur(null, null);
			}
			else if (mot.getValeur().equals(">"))
			{
				return new SuperieurStrict(null, null);
			}
			else if (mot.getValeur().equals("<="))
			{
				return new Inferieur(null, null);
			}
			else if (mot.getValeur().equals("<"))
			{
				return new InferieurStrict(null, null);
			}
			else if (mot.getValeur().equals("et"))
			{
				return new Et(null, null);
			}
			else if (mot.getValeur().equals("ou"))
			{
				return new Ou(null, null);
			}
			else if (mot.getValeur().equals("=="))
			{
				return new Egal(null, null);
			}
			else if (mot.getValeur().equals("!="))
			{
				return new Different(null, null);
			}
			else
			{
				throw new SyntaxException("L'indentificateur '" + mot.getValeur() +
								"' n'est pas valide.");
			}
		}
		else if (mot.getType() == MotLexical.TYPE_VARIABLE)
		{
			return new Reference(mot.getValeur(), null);
		}
		else if (mot.getType() == MotLexical.TYPE_AFFECTATION)
		{
			return new Affectation(null, null);
		}
		else if (mot.getType() == MotLexical.TYPE_FIN_INSTRUCTION)
		{
			return new PrincipaleExpression();
		}
		else if (mot.getType() == MotLexical.TYPE_NON)
		{
			return new Non(null);
		}
		else
		{
			throw new SyntaxException("L'indentificateur '" + mot.getValeur() +
							"' n'est pas valide.");
		}
	}

	/**
	 * Complète les expressions qui ne sont pas complète à gauche ou à droite
	 * @param listeExpressions La séquence d'expression
	 */
	private void completerLesExpressions(ArrayList<ExpressionAbstraite> listeExpressions) throws InvalidInstruction
	{
		ExpressionAbstraite expressionsNonSatisfaites[] =
						new ExpressionAbstraite[listeExpressions.size()];

		int ensIndex[] = new int[listeExpressions.size()];
		int nonSatisfaitesIndex = 0;

		// Boucle de transformation
		for (int i = 0; i < listeExpressions.size(); i++)
		{
			ExpressionAbstraite exp = listeExpressions.get(i);

			// On regarde si l'expression ext complete
			if (!exp.isCompleteAgauche() ||
							!exp.isCompleteAdroite())
			{
				expressionsNonSatisfaites[nonSatisfaitesIndex] = exp;
				ensIndex[nonSatisfaitesIndex] = i;
				nonSatisfaitesIndex++;
			}
		}

		// On fait un tri à bulles sur les expressions non satisfaites
		// en fonction de la priorité des opérateurs
		for (int i = 0; i < nonSatisfaitesIndex; ++i)
		{
			for (int j = i; j < nonSatisfaitesIndex; ++j)
			{
				if (expressionsNonSatisfaites[i].getPriorite() <
								expressionsNonSatisfaites[j].getPriorite())
				{
					ExpressionAbstraite temp = expressionsNonSatisfaites[i];
					expressionsNonSatisfaites[i] = expressionsNonSatisfaites[j];
					expressionsNonSatisfaites[j] = temp;
					int temp2 = ensIndex[i];
					ensIndex[i] = ensIndex[j];
					ensIndex[j] = temp2;
				}
			}
		}

		// Ce tri étant effectué, il va nous falloir satisfaire les expressions
		for (int i = 0; i < nonSatisfaitesIndex; ++i)
		{
			if (debug)
			{
				for (int j = 0; j < listeExpressions.size(); ++j)
				{
					System.out.println("[" + j + "] " + listeExpressions.get(j));
				}
			}

			ExpressionAbstraite exp = expressionsNonSatisfaites[i];

			if (debug)
			{
				System.out.println("On s'occupe de : [" + ensIndex[i] + "] " + exp);
			}

			completeExpression(exp, listeExpressions, ensIndex[i]);
		}

	}

	/**
	 * Complete l'expression donné en paramètre
	 * @param exp
	 * @param listeExpressions
	 * @param i
	 */
	private void completeExpression(ExpressionAbstraite exp,
					ArrayList<ExpressionAbstraite> listeExpressions, int expIndex) throws InvalidInstruction
	{
		if (!exp.isCompleteAgauche() && expIndex - 1 >= 0)
		{
			exp.setGauche(listeExpressions.get(expIndex - 1));
			ExpressionAbstraite oldExp = listeExpressions.get(expIndex - 1);
			int j = 1;
			while (expIndex - j >= 0 &&
							listeExpressions.get(expIndex - j) == oldExp)
			{
				listeExpressions.set(expIndex - j, exp);
				j++;
			}
		}
		if (!exp.isCompleteAdroite() && expIndex + 1 < listeExpressions.size())
		{
			ExpressionAbstraite oldExp = listeExpressions.get(expIndex + 1);
			exp.setDroite(listeExpressions.get(expIndex + 1));
			int j = 1;
			while (expIndex + j < listeExpressions.size() &&
							listeExpressions.get(expIndex + j) == oldExp)
			{
				listeExpressions.set(expIndex + j, exp);
				j++;
			}
		}
		if (!exp.isCompleteAgauche() || !exp.isCompleteAdroite())
		{
			System.out.println("Nous ne pouvons pas satisfaire : " + exp);
		}
	}

	public boolean isDebug()
	{
		return debug;
	}

	public void setDebug(boolean debug)
	{
		this.debug = debug;
	}
}
